# build.py

import os
import codecs
import pathlib
from setuptools import setup as _setup, find_namespace_packages
from typing import Iterable, Any

__all__ = [
    "Dependencies",
    "parse_requirements",
    "unfold",
    "get_dependencies",
    "build_manifest",
    "build_pyproject",
    "collect_files",
    "setup"
]

class Dependencies:

    def __init__(
            self,
            requirements: Iterable[str] = None,
            links: Iterable[str] = None
    ) -> None:

        if requirements is None:
            requirements = []

        if links is None:
            links = []

        self.requirements = requirements
        self.links = links

def parse_requirements(
        source: str | pathlib.Path,
        excluded: Iterable[str] = None
) -> Dependencies:

    if excluded is None:
        excluded = []

    links = []
    requirements = []

    with codecs.open(source, 'r') as requirements_txt:
        for line in requirements_txt.readlines():
            line = line.replace("\n", "")

            if not line or line.replace(" ", "").startswith("#"):
                continue

            line = line[:line.find("#")] if "#" in line else line

            requirement = line.split(";")

            for i, part in enumerate(parts := requirement[0].split()):
                name = parts[0]

                for char in ['=', '<', ">"]:
                    name = name[:name.find(char)] if char in name else name
                # end

                if name in excluded:
                    continue

                if (
                    (len(parts) > 2) and
                    (part in ("--extra-index-url", "--index-url", "--find-links"))
                ):
                    if i == 0:
                        name = parts[-1]

                    else:
                        name = parts[0]

                    if name in excluded:
                        continue

                    links.append(f"{parts[1]}#egg={name}")
                    requirements.append(";".join([name] + requirement[1:]))

                    break

                if len(parts) == 1 and name not in excluded:
                    requirements.append(line)

    return Dependencies(
        requirements=requirements,
        links=links
    )

def unfold(obj: Any, /, indentation: int = None) -> str:

    try:
        if isinstance(obj, str):
            raise TypeError

        string = str(obj)

        if not string:
            raise TypeError

        if indentation is None:
            indentation = 1

        return (
            string[0] + "\n" + ("\t" * indentation) +
            (
                (",\n" + ("\t" * indentation)).join(
                    (
                        unfold(element, indentation=indentation + 1)
                        for element in iter(obj)
                    )
                    if not isinstance(obj, dict) else
                    (
                        (
                            f"{unfold(key, indentation=indentation + 1)}: "
                            f"{unfold(value, indentation=indentation + 1)}"
                        )
                        for key, value in obj.items()
                    )
                )
            ) + "\n" + ("\t" * (indentation - 1)) +
            string[-1]
        )

    except TypeError:
        if isinstance(obj, bool):
            return repr(obj).lower()

        return repr(obj)
def get_dependencies(
        *,
        requirements: str | pathlib.Path = None,
        dev_requirements: str | pathlib.Path = None
) -> tuple[Dependencies, Dependencies]:

    if requirements is not None and os.path.exists(requirements):
        dependencies = parse_requirements(source=requirements)

        if dev_requirements is not None and os.path.exists(dev_requirements):
            extra_dependencies = parse_requirements(source=dev_requirements)

            with codecs.open(requirements, 'r') as req_file:
                reqs = req_file.read()

            with codecs.open(dev_requirements, 'r') as req_dev_file:
                dev_reqs = req_dev_file.read()

            if reqs not in dev_reqs:
                with codecs.open(dev_requirements, 'a') as req_dev_file:
                    req_dev_file.write(f"\n\n{reqs}")

        else:
            extra_dependencies = Dependencies()

    else:
        dependencies = Dependencies()
        extra_dependencies = Dependencies()

    return dependencies, extra_dependencies

def build_pyproject(
        project: str | pathlib.Path = None, **kwargs: Any
) -> None:

    if project is not None:
        content = "\n".join(
            [
                f"{key} = {unfold(value, indentation=1)}"
                for key, value in kwargs.items() if not (
                    isinstance(value, dict) or
                    (
                        key in (
                            "long_description",
                            "license",
                            'include_package_data',
                            'dependency_links',
                            'url',
                            'install_requires',
                            'extras_require',
                            'author',
                            'packages',
                            'author_email',
                            "data_files",
                            "long_description_content_type"
                        )
                    )
                )
            ]
        )

        if os.path.exists(project):
            with codecs.open(project, 'r') as project_file:
                project_content = project_file.read()

        else:
            project_content = ""

        if content not in project_content:
            with codecs.open(project, 'w') as req_dev_file:
                req_dev_file.write(
                    '[build-system]\n'
                    'requires = ["setuptools"]\n'
                    'build-backend = "setuptools.build_meta"\n'
                    f'\n[project]\n{content}'
                )

def build_manifest(
        include: Iterable[str] = None, manifest: bool = None
) -> None:

    if not os.path.exists("MANIFEST.in") or not manifest:
        with codecs.open("MANIFEST.in", 'w') as include_file:
            include_file.write("")

    if os.path.exists("MANIFEST.in"):
        with codecs.open("MANIFEST.in", 'r') as include_file:
            include_content = include_file.read()

        with codecs.open("MANIFEST.in", 'a') as include_file:
            for line in include:
                if line not in include_content:
                    include_file.write(f"include {line}\n")

def collect_files(location: str | pathlib.Path, levels: int = None) -> list[str]:

    paths = []

    if levels == 0:
        return paths

    for name in os.listdir(location):
        path = pathlib.Path(location) / pathlib.Path(name)

        if path.is_file():
            paths.append(str(path))

        elif path.is_dir():
            paths.extend(
                collect_files(
                    str(path), levels=(
                        levels - 1 if levels is not None else levels
                    )
                )
            )

    return paths

def setup(
        package: str | pathlib.Path, *,
        readme: str | bool | pathlib.Path = None,
        exclude: Iterable[str | pathlib.Path] = None,
        include: Iterable[str | pathlib.Path] = None,
        requirements: str | pathlib.Path = None,
        dev_requirements: str | pathlib.Path = None,
        project: str | pathlib.Path = None,
        manifest: bool = None,
        **kwargs: Any
) -> None:

    if include is None:
        include = []

    else:
        include = list(include)

    if readme in (None, True):
        readme = 'README.md' if os.path.exists('README.md') else None

    if exclude is None:
        exclude = ()

    if isinstance(readme, (str, pathlib.Path)):
        with codecs.open(str(readme), 'r') as desc_file:
            long_description = desc_file.read()

    else:
        long_description = None

    dependencies, extra_dependencies = get_dependencies(
        requirements=requirements, dev_requirements=dev_requirements
    )

    include = list(
        set(
            include + [
                file for file in
                [requirements, dev_requirements, project, "build.py"]
                if file and os.path.exists(str(file))
            ]
        )
    )

    for included in list(include):
        if included is not None and os.path.isdir(str(included)):
            include += list(set(collect_files(location=included)))

    include = [str(pathlib.Path(path)) for path in include]

    packages = [
        f"{'.'.join(pathlib.Path(str(package)).parts)}.{content}"
        for content in find_namespace_packages(package, exclude=exclude)
    ] + [package]

    kwargs.setdefault("extras_require", {}).setdefault(
        "dev", [
            req for req in extra_dependencies.requirements
            if req not in dependencies.requirements
        ]
    )

    build_manifest(include=include, manifest=manifest)

    if project:
        build_pyproject(project=project, **kwargs)

    _setup(
        **(dict(project=project) if project else {}),
        packages=packages,
        long_description=long_description,
        include_package_data=True,
        dependency_links=dependencies.links,
        install_requires=dependencies.requirements,
        **kwargs
    )



# sniff.py

from io import BytesIO
import threading

from scapy.all import (
    PacketList, wrpcap, Packet, cast, rdpcap,
    AsyncSniffer as ScapyAsyncSniffer
)

from sniffingio.data import SniffSettings
from sniffingio.filters import format_packet_filters, load_packet_filter

__all__ = [
    "sniff",
    "Sniffer",
    "write_pcap",
    "read_pcap"
]

class Sniffer:

    def __init__(self, data: SniffSettings = None) -> None:

        if data is None:
            data = SniffSettings()

        self.data = data

        self._sniffer = ScapyAsyncSniffer()

    def packets(self) -> PacketList:

        return cast(PacketList, self._sniffer.results)

    def start(self, data: SniffSettings = None) -> PacketList:

        data = data or self.data or SniffSettings()

        callback = None

        if data.callback and data.printer:
            callback = lambda p: (data.callback(p), data.printer(p))

        elif data.callback:
            callback = data.callback

        elif data.printer:
            if data.printer is True:
                callback = print

            else:
                callback = data.printer

        static_filter = None

        if data.static_filter is not None:
            static_filter = data.static_filter

            if isinstance(static_filter, dict):
                static_filter = load_packet_filter(static_filter)

            if not isinstance(static_filter, str):
                static_filter = format_packet_filters(static_filter)

        # noinspection PyProtectedMember
        self._sniffer._run(
            count=data.count,
            store=data.store,
            quiet=data.quiet,
            timeout=data.timeout,
            iface=data.interface,
            prn=callback,
            lfilter=data.live_filter,
            filter=static_filter,
            stop_filter=data.stop_filter,
            started_callback=data.start_callback
        )

        return self.packets()

    def thread_start(self) -> None:

        threading.Thread(target=self.start).start()

    def stop(self) -> None:

        self._sniffer.stop()

def sniff(data: SniffSettings) -> PacketList:

    return Sniffer(data).start()

def write_pcap(packet: PacketList | list[Packet] | Packet, io: BytesIO | str) -> None:

    wrpcap(io, packet)

def read_pcap(io: BytesIO | str) -> PacketList | list[Packet] | Packet:

    return rdpcap(io)



# callbacks.py

from io import BytesIO
from dataclasses import dataclass
from typing import Callable

from scapy.all import Packet, PacketList, wrpcap, rdpcap

__all__ = [
    "dump_packet",
    "load_packet",
    "PacketCallback"
]

def dump_packet(packet: Packet | PacketList) -> bytes:

    data = BytesIO()

    payload = []

    data.write = payload.append

    wrpcap(data, packet)

    return b"".join(payload)

def load_packet(data: bytes) -> PacketList:

    return rdpcap(BytesIO(data))

@dataclass(slots=True)
class PacketCallback:

    callback: Callable[[Packet], ...]

    disabled: bool = False
    result: bool = False

    def __call__(self, *args, **kwargs) -> ...:

        return self.execute(*args, **kwargs)

    def disable(self) -> None:

        self.disabled = True

    def enable(self) -> None:

        self.disabled = False

    def execute(self, packet: Packet) -> ...:

        if self.disabled:
            return

        result = self.callback(packet)

        if not self.result:
            return

        return result



# data.py

from dataclasses import dataclass
from typing import Callable, Iterable

from sniffingio.callbacks import PacketCallback
from sniffingio.filters import LivePacketFilter, PacketFilterOperand

from scapy.all import NetworkInterface, PacketList, Packet

__all__ = [
    "SniffSettings",
    "NetworkInterface",
    "Packet",
    "PacketList"
]

@dataclass(slots=True)
class SniffSettings:

    count: int = 0
    timeout: int = None
    store: bool = True
    quiet: bool = True
    callback: PacketCallback = None
    printer: bool | PacketCallback = None
    live_filter: LivePacketFilter = None
    stop_filter: LivePacketFilter = None
    interface: str | NetworkInterface = None
    static_filter: str | dict | PacketFilterOperand | Iterable[PacketFilterOperand] = None
    start_callback: Callable[[], ...] = None



# filters.py

from typing import Iterable, Callable, ClassVar, Self
from dataclasses import dataclass, asdict
from abc import ABCMeta, abstractmethod

from dacite import from_dict

from scapy.all import Packet

__all__ = [
    "PacketFilterOperand",
    "PacketFilter",
    "PacketFilterIntersection",
    "PacketFilterOperator",
    "PacketFilterUnion",
    "PacketFilterNegation",
    "BasePacketFilterUnion",
    "BasePacketFilter",
    "StaticPacketFilter",
    "BasePacketFilterOperator",
    "BasePacketFilterIntersection",
    "format_packet_filters",
    "LivePacketFilter",
    "dump_packet_filter",
    "load_packet_filter",
    "PacketFilterValues"
]

def wrap(value: str) -> str:

    if (" " in value) and not (value.startswith("(") and value.endswith(")")):
        value = f"({value})"

    return value

class BasePacketFilterOperator(metaclass=ABCMeta):

    @staticmethod
    def format_join(values: Iterable[str], joiner: str) -> str:

        if not values:
            return ""

        values = tuple(str(value) for value in values)

        if len(values) == 1:
            return wrap(values[0])

        data = f" {joiner} ".join(wrap(value) for value in values if value)

        return f"({data})"

class BasePacketFilterUnion(BasePacketFilterOperator, metaclass=ABCMeta):

    @classmethod
    def format_union(cls, values: Iterable[str]) -> str:

        return cls.format_join(values, joiner="or")

class BasePacketFilterIntersection(BasePacketFilterOperator, metaclass=ABCMeta):

    @classmethod
    def format_intersection(cls, values: Iterable[str]) -> str:

        return cls.format_join(values, joiner="and")

class BasePacketFilter(
    BasePacketFilterUnion,
    BasePacketFilterIntersection,
    metaclass=ABCMeta
):

    @abstractmethod
    def format(self) -> str:

        return ""

@dataclass(slots=True, frozen=True)
class PacketFilterOperand(BasePacketFilter, metaclass=ABCMeta):

    def __invert__(self) -> "PacketFilterOperand":

        if isinstance(self, PacketFilterNegation):
            return self.filter

        return PacketFilterNegation(self)

    def __or__(self, other: ...) -> "PacketFilterUnion":

        if isinstance(other, PacketFilterOperand):
            filters = []

            if isinstance(self, PacketFilterUnion):
                filters.extend(self.filters)

            else:
                filters.append(self)

            if isinstance(other, PacketFilterUnion):
                filters.extend(other.filters)

            else:
                filters.append(other)

            return PacketFilterUnion(tuple(filters))

        return NotImplemented

    def __and__(self, other: ...) -> "PacketFilterIntersection":

        if isinstance(other, PacketFilterOperand):
            filters = []

            if isinstance(self, PacketFilterIntersection):
                filters.extend(self.filters)

            else:
                filters.append(self)

            if isinstance(other, PacketFilterIntersection):
                filters.extend(other.filters)

            else:
                filters.append(other)

            return PacketFilterIntersection(tuple(filters))

        return NotImplemented

    @classmethod
    def load(cls, data: dict[str, ...]) -> Self:

        return from_dict(cls, data)

    def dump(self) -> dict[str, ...]:

        return asdict(self)

@dataclass(slots=True, frozen=True)
class StaticPacketFilter(PacketFilterOperand):

    filter: str

    def format(self) -> str:

        return self.filter

@dataclass(slots=True, frozen=True)
class PacketFilterOperator(PacketFilterOperand, metaclass=ABCMeta):

    filters: tuple["PacketFilterOperand", ...]

    def __len__(self) -> int:

        return len(self.filters)

@dataclass(slots=True, frozen=True)
class PacketFilterUnion(PacketFilterOperator, BasePacketFilterUnion):

    def format(self) -> str:

        return self.format_union(
            (f.format() for f in self.filters or ())
        )

@dataclass(slots=True, frozen=True)
class PacketFilterIntersection(PacketFilterOperator, BasePacketFilterIntersection):

    def format(self) -> str:

        return self.format_intersection(
            (f.format() for f in self.filters or ())
        )

@dataclass(slots=True, frozen=True)
class PacketFilterNegation(PacketFilterOperand):

    filter: PacketFilterOperand

    ATTRIBUTES: ClassVar[set[str]] = {'filter'}

    def format(self) -> str:

        data = self.filter.format()

        if not data:
            return ""

        return f"(not {data})"

def layers_names(packet: Packet) -> list[str]:

    names = [packet.name]

    while packet.payload:
        packet = packet.payload

        names.append(packet.name)

    return names

@dataclass(slots=True, frozen=True)
class PacketFilterValues[T](PacketFilterOperand):

    types: list[str] = None
    names: list[str] = None
    values: list[T] = None
    source_values: list[T] = None
    destination_values: list[T] = None

    @classmethod
    def load(cls, data: dict[str, list[T]]) -> "PacketFilterValues[T]":

        return from_dict(cls, data)

    def dump(self) -> dict[str, list[T]]:

        return asdict(self)

    @classmethod
    def format_values(cls, values: Iterable[str], key: str = None) -> str:

        if not values:
            return ""

        return cls.format_union(
            (
                " ".join((key, str(value)) if key else (str(value),))
                for value in values
                if value
            )
        )

    def format(self) -> str:

        values = [
            self.format_union(values)
            for values in (
                self.types,
                (
                    self.format_values(self.values, key=name)
                    for name in self.names or ['']
                ),
                (
                    self.format_values(
                        self.source_values, key=' '.join(['src', name])
                    )
                    for name in self.names or ['']
                ),
                (
                    self.format_values(
                        self.destination_values, key=' '.join(['dst', name])
                    )
                    for name in self.names or ['']
                )
            )
            if values
        ]

        values = [value for value in values if value]

        return self.format_intersection(values)

@dataclass(slots=True, frozen=True, eq=False)
class PacketFilter(PacketFilterOperand):

    data_link: PacketFilterValues[str] = None
    internet: PacketFilterValues[str] = None
    transportation: PacketFilterValues[int] = None

    def match(self, packet: Packet) -> bool:

        layers = (self.data_link, self.internet, self.transportation)

        for layer, layer_filter in zip(packet.layers(), layers):
            if layer_filter is None:
                continue

            layer_filter: PacketFilterValues

            if layer.name.lower() not in {n.lower() for n in layer_filter.types}:
                return False

            if hasattr(layer, 'src'):
                src = layer.src
                dst = layer.dst

            elif hasattr(layer, 'sport'):
                src = layer.sport
                dst = layer.dport

            else:
                continue

            sources = layer_filter.values + layer_filter.source_values
            destinations = layer_filter.values + layer_filter.destination_values

            if (
                (sources and (src not in sources)) or
                (destinations and (dst not in destinations))
            ):
                return False

        return True

    def format(self) -> str:

        return self.format_intersection(
            (
                self.data_link.format(),
                self.internet.format(),
                self.transportation.format()
            )
        )

def format_packet_filters(
        filters: BasePacketFilter | Iterable[BasePacketFilter],
        joiner: PacketFilterOperator = PacketFilterUnion
) -> str:

    if joiner is None:
        joiner = PacketFilterUnion

    if isinstance(filters, PacketFilterOperand):
        return filters.format()

    return joiner(tuple(filters)).format()

@dataclass(slots=True)
class LivePacketFilter:

    validator: Callable[[Packet], bool]

    disabled: bool = False

    def __call__(self, *args, **kwargs) -> bool:

        return self.validate(*args, **kwargs)

    def disable(self) -> None:

        self.disabled = True

    def enable(self) -> None:

        self.disabled = False

    def validate(self, packet: Packet) -> bool:

        if self.disabled:
            return True

        result = self.validator(packet)

        return result

PF = (
    PacketFilter |
    PacketFilterUnion |
    PacketFilterIntersection |
    PacketFilterNegation |
    StaticPacketFilter
)

def dump_packet_filter(data: PF) -> dict[str, ...]:

    return data.dump()

def load_packet_filter(data: PF | str | dict[str, ...]) -> PF:

    if isinstance(data, PacketFilterOperand):
        return data

    if isinstance(data, str):
        return StaticPacketFilter(data)

    return PacketFilterOperand.load(data)
